import logging
import threading
import unittest

from hpfeeds.blocking import ClientSession

from .fakebroker import FakeBroker, setup_default_reactor


class TestClientIntegration(unittest.TestCase):

    log = logging.getLogger('hpfeeds.testserver')

    def _server_thread(self):
        self.reactor = setup_default_reactor(self)
        self.server.start()
        self.reactor.run(installSignalHandlers=False)

    def setUp(self):
        self.server = FakeBroker()

        self.server_thread = threading.Thread(
            target=self._server_thread,
        )
        self.server_thread.start()

    def test_subscribe_and_publish(self):
        with ClientSession('127.0.0.1', self.server.port, 'test', 'secret') as c:
            c.subscribe('test-chan')
            c.publish('test-chan', b'data')

            for ident, channel, payload in c:
                assert payload == b'data'
                assert channel == 'test-chan'
                assert ident == 'test'
                break

            c.unsubscribe('test-chan')

    def tearDown(self):
        self.log.debug('Cancelling future')
        self.server.close()
        self.reactor.callFromThread(self.reactor.stop)
        self.log.debug('Waiting')
        self.server_thread.join()

        assert len(self.server.connections) == 0, 'Connection left dangling'



from __future__ import absolute_import

import logging
import socket
import sys

from hpfeeds.protocol import msgpublish, msgsubscribe, msgunsubscribe

from .protocol import ClientProtocol
from .queue import Queue
from .reactor import ThreadReactor


class Protocol(ClientProtocol):

    def __init__(self, session):
        self.session = session
        super(Protocol, self).__init__(session.ident, session.secret)

    def on_publish(self, ident, channel, payload):
        self.session.read_queue.put_nowait((ident, channel, payload))


class ClientSession(object):

    '''
    Create and maintain a session with the hpfeeds broker.

    The connection is managed in a dedicated thread.
    '''

    def __init__(self, host, port, ident, secret):
        self.host = host
        self.port = port
        self.ident = ident
        self.secret = secret
        self.subscriptions = set()

        self.read_queue = Queue()

        self._reactor = ThreadReactor(self._build_protocol, self._connect)

    def _build_protocol(self):
        return Protocol(self)

    def _connect(self):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.host, self.port))
        except Exception as e:
            logging.exception(e)
            return None
        return sock

    def start(self):
        self._reactor.start()

    def stop(self):
        self._reactor.stop()

    def publish(self, channel, payload):
        self._reactor.write(msgpublish(self.ident, channel, payload))

    def publish_iter(self, channel, iterator):
        for payload in iterator:
            self.publish(channel, payload)

    def subscribe(self, channel):
        self.subscriptions.add(channel)
        self._reactor.write(msgsubscribe(self.ident, channel))

    def unsubscribe(self, channel):
        self.subscriptions.discard(channel)
        self._reactor.write(msgunsubscribe(self.ident, channel))

    def read(self):
        return self.read_queue.get()

    def __enter__(self):
        self.start()
        self._reactor.when_connected.wait()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def __iter__(self):
        return self

    def __next__(self):
        if self._reactor.closing:
            raise StopIteration()
        return self.read()

    if sys.version_info[0] == 2:
        def next(self):
            return self.__next__()



import asyncio
import socket
import unittest

import aiohttp

from hpfeeds.asyncio.client import ClientSession
from hpfeeds.broker import prometheus
from hpfeeds.broker.auth.memory import Authenticator
from hpfeeds.broker.server import Server


class TestBrokerPrometheusEndpoint(unittest.TestCase):

    def setUp(self):
        prometheus.reset()

        authenticator = Authenticator({
            'test': {
                'secret': 'secret',
                'subchans': ['test-chan'],
                'pubchans': ['test-chan'],
                'owner': 'some-owner',
            }
        })

        self.server = Server(authenticator, bind='127.0.0.1:20000', exporter='127.0.0.1:20001')

    def test_metrics_server(self):
        async def inner():
            server_future = asyncio.ensure_future(self.server.serve_forever())
            await self.server.when_started

            async with aiohttp.ClientSession() as session:
                async with session.get('http://127.0.0.1:20001/metrics') as resp:
                    metrics = await resp.text()
                    print(metrics)
                    assert 'hpfeeds_broker_client_connections 0.0' in metrics
                    assert 'hpfeeds_broker_connection_send_buffer_size{' not in metrics

            sock = socket.socket()
            sock.connect(('127.0.0.1', 20000))

            async with aiohttp.ClientSession() as session:
                async with session.get('http://127.0.0.1:20001/metrics') as resp:
                    metrics = await resp.text()
                    print(metrics)
                    assert 'hpfeeds_broker_client_connections 1.0' in metrics
                    assert 'hpfeeds_broker_connection_send_buffer_size{' not in metrics

            sock.close()

            async with ClientSession('127.0.0.1', 20000, 'test', 'secret'):
                async with aiohttp.ClientSession() as session:
                    async with session.get('http://127.0.0.1:20001/metrics') as resp:
                        metrics = await resp.text()
                        print(metrics)
                        assert 'hpfeeds_broker_client_connections 1.0' in metrics
                        assert 'hpfeeds_broker_connection_send_buffer_size{ident="test"} 0.0' in metrics

            server_future.cancel()
            await server_future

        asyncio.get_event_loop().run_until_complete(inner())



from __future__ import absolute_import

import os
import socket

try:
    import queue
except ImportError:
    import Queue as queue

__all__ = [
    'Empty',
    'Queue',
]


Empty = queue.Queue


class Queue(queue.Queue, object):

    def __init__(self):
        super(Queue, self).__init__()
        if os.name == 'posix':
            self._putsocket, self._getsocket = socket.socketpair()
        else:
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server.bind(('127.0.0.1', 0))
            server.listen(1)
            self._putsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._putsocket.connect(server.getsockname())
            self._getsocket, _ = server.accept()
            server.close()

    def fileno(self):
        return self._getsocket.fileno()

    def put(self, item, block=True):
        super(Queue, self).put(item, block)
        self._putsocket.send(b'x')

    def get(self, block=True):
        self._getsocket.recv(1)
        return super(Queue, self).get(block)



import select
import unittest

from hpfeeds.blocking.queue import Queue


class TestQueuePollable(unittest.TestCase):

    def test_write_queue(self):
        q = Queue()

        q.put_nowait('some data')

        # Queue has some data in - so should be readable
        r, w, x = select.select([q], [], [], 0)
        assert q in r

        # Queue still has data in, so still be readable
        r, w, x = select.select([q], [], [], 0)
        assert q in r

        assert q.get_nowait() == 'some data'

        # Queue has no data in, so no longer be readable
        r, w, x = select.select([q], [], [], 0)
        assert q not in r

    '''
    @mock.patch.object(os, 'name', new_callable=mock.PropertyMock)
    def test_fallback(self, os_name):
        os_name.return_value = 'nt'

        q = Queue()

        q.put_nowait('some data')

        # Queue has some data in - so should be readable
        r, w, x = select.select([q], [], [], 0)
        assert q in r

        # Queue still has data in, so still be readable
        r, w, x = select.select([q], [], [], 0)
        assert q in r

        assert q.get_nowait() == 'some data'

        # Queue has no data in, so no longer be readable
        r, w, x = select.select([q], [], [], 0)
        assert q not in r
    '''



'''
This module implements an incredibly basic broker.

In has no authentication and no channels.

It is meant to support py 2 and py 3.
'''

import socket
import sys

from twisted.internet.protocol import Factory

from hpfeeds.twisted import BaseProtocol


class FakeBrokerProtocol(BaseProtocol):

    def connectionMade(self):
        self.info('hpfeeds', b'\x00' * 8)
        self.factory.broker.connections.add(self)

    def connectionLost(self, reason):
        self.factory.broker.connections.discard(self)

    def onAuth(self, ident, chan):
        pass

    def onSubscribe(self, ident, chan):
        pass

    def onUnsubscribe(self, ident, chan):
        pass

    def onPublish(self, ident, chan, payload):
        for con in list(self.factory.broker.connections):
            con.publish(ident, chan, payload)


class FakeBroker(object):

    def __init__(self):
        self.connections = set()

        self.sock = sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setblocking(False)
        sock.bind(('127.0.0.1', 0))
        sock.listen(100)

        self.port = sock.getsockname()[1]

    def start(self):
        factory = Factory.forProtocol(FakeBrokerProtocol)
        factory.broker = self

        from twisted.internet import reactor
        self.adopted_port = reactor.adoptStreamPort(self.sock.fileno(), socket.AF_INET, factory)
        print('Port adopted', self.port)

    def close(self):
        return self.adopted_port.stopListening()


def setup_default_reactor(test):
    if 'twisted.internet.reactor' in sys.modules:
        del sys.modules['twisted.internet.reactor']

    from twisted.internet import selectreactor
    selectreactor.install()

    from twisted.internet import reactor
    return reactor


def setup_asyncio_reactor(test):
    import asyncio
    from twisted.internet import asyncioreactor

    _old_loop = asyncio.get_event_loop()

    if 'twisted.internet.reactor' in sys.modules:
        del sys.modules['twisted.internet.reactor']

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    asyncioreactor.install(eventloop=loop)

    def cleanup():
        asyncio.get_event_loop().close()
        asyncio.set_event_loop(_old_loop)

    test.addCleanup(cleanup)

    from twisted.internet import reactor
    return reactor



import sys


collect_ignore = []
if sys.version_info[0] == 2:
    collect_ignore.extend([
        "hpfeeds/tests/test_asyncio_client.py",
        "hpfeeds/tests/test_asyncio_protocol.py",
        "hpfeeds/tests/test_broker_connection.py",
        "hpfeeds/tests/test_broker_prometheus.py",
        "hpfeeds/tests/test_client_integration.py",
        "hpfeeds/tests/test_twisted_service.py",
    ])




from __future__ import absolute_import

import errno
import logging
import select
import socket
import sys
import threading

from . import queue


class Reactor(object):

    '''
    This class handles IO multiplexing and connection failure handling.
    '''

    def __init__(self, protocol_class, connector):
        self.protocol_class = protocol_class
        self.connector = connector

        self.closing = False
        self.when_connected = threading.Event()

        self._outbox = queue.Queue()

    def write(self, data):
        self._outbox.put_nowait(data)

    def close(self):
        self.sock.close()

    def _connect(self):
        '''
        Establish a connection using ``self.connector`` and then set up
        internal state for that new connection.
        '''
        self.sock = self.connector()

        self.sock.setblocking(False)

        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if sys.platform.startswith('linux'):
            self.sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPIDLE, 10)
            self.sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPINTVL, 5)
            self.sock.setsockopt(socket.SOL_TCP, socket.TCP_KEEPCNT, 3)

        self._buffer = b''
        self._outbox = queue.Queue()

        self.protocol = self.protocol_class()
        self.protocol.transport = self
        self.protocol.connection_made()

        self.when_connected.set()

    def _select(self):
        '''
        Block (and don't consume CPU) until self.sock or self._outbox is ready
        for us
        '''
        logging.debug('_select')

        want_read = [self.sock]
        want_write = []

        if self._buffer:
            want_write.append(self.sock)
        else:
            want_read.append(self._outbox)

        r, w, x = select.select(want_read, want_write, [])

        if self.sock in r and not self._socket_read_ready():
            return

        if self._outbox in r and not self._outbox_read_ready():
            return

        if self.sock in w and not self._socket_write_ready():
            return

    def _socket_read_ready(self):
        logging.debug('_socket_ready_ready')
        try:
            data = self.sock.recv(1024)
        except socket.error as e:
            # Interupted by.. interupt, try again
            if e.args[0] == errno.EAGAIN:
                return True

            # Would block, so return and go back to sleep
            if e.args[0] == errno.EWOULDBLOCK:
                return True

            raise

        if not data:
            self._connection_lost('Read failed')
            return False

        self.protocol.data_received(data)
        return True

    def _outbox_read_ready(self):
        logging.debug('_outbox_read_ready')

        try:
            self._buffer += self._outbox.get_nowait()
        except socket.error as e:
            # Interupted by.. interupt, try again
            if e.args[0] == errno.EAGAIN:
                return True

            # Would block, so return and go back to sleep
            if e.args[0] == errno.EWOULDBLOCK:
                return True
        except queue.Empty:
            # Someone lied - there is no data really
            # Go to sleep and try again later
            return True

        # If we were blocked waiting for something to land in queue there is
        # good chance the write socket is ready. Try to send.
        self._socket_write_ready()

    def _socket_write_ready(self):
        logging.debug('_socket_write_ready')
        try:
            sent = self.sock.send(self._buffer)
        except socket.error as e:
            # Interupted by.. interupt, try again
            if e.args[0] == errno.EAGAIN:
                return True

            # Would block, so return and go back to sleep
            if e.args[0] == errno.EWOULDBLOCK:
                return True

            raise

        if sent == 0:
            self._connection_lost('Write failed')
            return False

        self._buffer = self._buffer[sent:]
        return True

    def _connection_lost(self, reason):
        self.protocol.connection_lost(reason)
        try:
            self.sock.close()
        except Exception:
            pass
        self.sock = None
        self.when_connected.clear()

    def run_forever(self):
        while not self.closing:
            self._connect()

            while not self.closing and self.sock:
                self._select()

    def stop(self):
        self.closing = True
        # if self.sock:
        #     self._connection_lost('Client session terminated')


class ThreadReactor(Reactor):

    def start(self):
        self._thread = threading.Thread(target=self.run_forever)
        self._thread.start()

    def stop(self):
        super(ThreadReactor, self).stop()
        self._thread.join()



from __future__ import print_function

import unittest

from twisted.internet import defer
from twisted.internet.endpoints import HostnameEndpoint

from hpfeeds.twisted import ClientSessionService

from .fakebroker import FakeBroker, setup_default_reactor


class TestClientIntegration(unittest.TestCase):

    def setUp(self):
        self.reactor = setup_default_reactor(self)

        self.server = FakeBroker()
        self.server.start()

    def test_subscribe_and_publish(self):
        @defer.inlineCallbacks
        def inner(reactor):
            print('Creating client service')
            endpoint = 'tcp:127.0.0.1:{}'.format(self.server.port)
            client = ClientSessionService(endpoint, 'test', 'secret')
            client.subscribe('test-chan')

            print('Starting client service')
            client.startService()

            # Wait till client connected
            print('Waiting to be connected')
            yield client.whenConnected

            print('Publishing test message')
            client.publish('test-chan', b'test message')

            print('Waiting for read()')
            payload = yield client.read()
            assert ('test', 'test-chan', b'test message') == payload

            print('Stopping client')
            yield client.stopService()

            print('Stopping server for reals')
            yield self.server.close()

        inner(self.reactor).addBoth(lambda *x: self.reactor.stop())
        self.reactor.run()

    def test_subscribe_and_publish_endpoint_impl(self):
        @defer.inlineCallbacks
        def inner(reactor):
            print('Creating client service')

            from twisted.internet import reactor
            endpoint = HostnameEndpoint(reactor, '127.0.0.1', self.server.port)

            client = ClientSessionService(endpoint, 'test', 'secret')
            client.subscribe('test-chan')

            print('Starting client service')
            client.startService()

            # Wait till client connected
            print('Waiting to be connected')
            yield client.whenConnected

            print('Publishing test message')
            client.publish('test-chan', b'test message')

            print('Waiting for read()')
            payload = yield client.read()
            assert ('test', 'test-chan', b'test message') == payload

            print('Stopping client')
            yield client.stopService()

            print('Stopping server for reals')
            yield self.server.close()

        inner(self.reactor).addBoth(lambda *x: self.reactor.stop())
        self.reactor.run()

    def test_subscribe_and_publish_endpoint_invalid(self):
        self.assertRaises(ValueError, ClientSessionService, 777, 'test', 'secret')



from __future__ import absolute_import

from hpfeeds.exceptions import ProtocolException
from hpfeeds.protocol import (
    OP_AUTH,
    OP_ERROR,
    OP_INFO,
    OP_PUBLISH,
    OP_SUBSCRIBE,
    OP_UNSUBSCRIBE,
    Unpacker,
    msgauth,
    readauth,
    readerror,
    readinfo,
    readpublish,
    readsubscribe,
    readunsubscribe,
)


class BaseProtocol(object):

    def __init__(self):
        self.unpacker = Unpacker()

    def connection_made(self):
        pass

    def connection_lost(self, reason):
        pass

    def protocol_error(self, reason):
        pass

    def on_error(self, error):
        raise NotImplementedError(self.on_info)

    def on_info(self, name, rand):
        raise NotImplementedError(self.on_info)

    def on_auth(self, ident, hash):
        raise NotImplementedError(self.on_auth)

    def on_publish(self, ident, chan, data):
        raise NotImplementedError(self.on_publish)

    def on_subscribe(self, ident, channel):
        raise NotImplementedError(self.on_subscribe)

    def on_unsubscribe(self, ident, channel):
        raise NotImplementedError(self.on_unsubscribe)

    def message_received(self, opcode, data):
        if opcode == OP_ERROR:
            return self.on_error(readerror(data))
        elif opcode == OP_INFO:
            return self.on_info(*readinfo(data))
        elif opcode == OP_AUTH:
            return self.on_auth(*readauth(data))
        elif opcode == OP_PUBLISH:
            return self.on_publish(*readpublish(data))
        elif opcode == OP_SUBSCRIBE:
            return self.on_subscribe(*readsubscribe(data))
        elif opcode == OP_UNSUBSCRIBE:
            return self.on_unsubscribe(*readunsubscribe(data))

        # Can't recover from an unknown opcode, so drop connection
        self.protocol_error('Unknown message opcode: {!r}'.format(opcode))
        self.transport.close()

    def data_received(self, data):
        self.unpacker.feed(data)
        try:
            for opcode, data in self.unpacker:
                self.message_received(opcode, data)
        except ProtocolException as e:
            self.protocol_error(str(e))
            self.transport.close()


class ClientProtocol(BaseProtocol):

    def __init__(self, ident, secret):
        self.ident = ident
        self.secret = secret
        super(ClientProtocol, self).__init__()

    def connection_ready(self):
        ''' Called when a client has connected to broker and sent authentication '''
        pass

    def on_info(self, name, rand):
        self.transport.write(msgauth(rand, self.ident, self.secret))
        self.connection_ready()

    def on_auth(self, ident, hash):
        self.protocol_error('Unexpected OP_AUTH')
        self.transport.close()

    def on_subscribe(self, ident, channel):
        self.protocol_error('Unexpected OP_SUBSCRIBE')
        self.transport.close()

    def on_unsubscribe(self, ident, channel):
        self.protocol_error('Unexpected OP_UNSUBSCRIBE')
        self.transport.close()
